(*
\section{The Lexer} 

\subsection{Declarations}

*)
{

(*

  This module contructs lexical entities and transmits them to the HLPSL parser.
  The keywords are stored in a hashtable. This solution permits to have
  a smaller automaton (see Ocaml Manual 12.7, comon errors). The tokens are
  defined in the \texttt{parser.mly} file.
  \label{lexer:keyword:table}
*)

open Parser;;
exception Eof;;

let keyword_table = Hashtbl.create 68;;


(*
  We now put all reserved word into a hash table, and define functions 
  to access this hash table.
*)

List.iter
  (fun (keyword,token) ->
    Hashtbl.add keyword_table keyword token)
(*
  The following list contains the pairs (keyword,token). For new Ocaml 
  users, note that the tuple constructor is the comma, and the list
  constructor is the semi-colon.
*)
  [
   "accept",ACCEPT;
   "agent",AGENT;
   "authenticates", AUTHENTICATES;
   "authentication_on", AUTHENTICATION_ON;
   "bool",BOOL;
   "channel",CHANNEL;
   "composition",COMPOSITION;
   "cons",CONS;
   "const",CONST;
   "delete",DELETE;
   "dy",DY;
   "end",END;
   "exists",LOCAL;
   "goal",GOAL;
   "hash", HASH;
   "function", HASH_FUNC;
   "hash_func", HASH_FUNC;
   "in",IN;
   "init",INIT;
   "intruder_knowledge",IKNOWLEDGE;
   "inv",INV;
(*   "list",LIST; *)
   "local",LOCAL;
(*   "loop",LOOP;*)
   "message",MESSAGE;
   "nat",NAT;
   "new",NEW;
   "not",NOT;
   "on", ON;
   "ota",OTA;
   "owns",OWNS;
   "played_by",PLAYED_BY;
   "protocol_id",PROTOCOL_ID;
   "public_key",PUBLIC_KEY;
   "role",ROLE;
   "secrecy_of",SECRECY_OF;
   "set",SET;
   "start",START;
   "symmetric_key",SYMMETRIC_KEY;
   "text",TEXT;
   "transition",TRANSITION;
   "weakly",WEAKLY;
   "weak_authentication_on", WEAK_AUTHENTICATION_ON;
 ]
;;


(*
  The two next functions are not optimal, as a keyword will often be
  put twice in uppercase. However, as long as we have no stable
  version, it seems better to ensure that keywords are matched in
  uppercase. 
*)

let is_keyword identifier =
  Hashtbl.mem keyword_table (String.lowercase identifier);;

let get_keyword identifier =
  Hashtbl.find keyword_table (String.lowercase identifier);;
  
(*
  We think this is a good idea to prefix all variables and constants
  in order to avoid confusion with the IF. It could be in principle
  possible to forbid the use of confusing identifiers, but this would
  imply the end-user also have to know the IF, and we do not want that. 

  Thus, we prefix identifiers with the value of
  [identifierPrefix]. We use the [prefixName] fonction to prefix only
  those identifiers which will lead to confusion (and thus permits a
  better readability of the IF in most
  cases.)\label{lexer:identifier:prefix} \label{lexer:prefix:name}
*)

let identifier_prefix = "id";;
  
let prefix_name name =
  match name with
  | "apply"
  | "attack_state"
  | "attack_states"
  | "contains"
  | "crypt"
  | "dummy_agent"
  | "dummy_bool"
  | "dummy_chnl"
  | "dummy_chnl_dy"
  | "dummy_chnl_ota"
  | "dummy_hash"
  | "dummy_msg"
  | "dummy_nat"
  | "dummy_nonce"
  | "dummy_pk"
  | "dummy_set"
  | "dummy_sk"
  | "equal"
  | "equations"
  | "function"
  | "initial_state"
  | "inits"
  | "intruder"
  | "leq"
  | "pair"
  | "properties"
  | "property"
  | "rules"
  | "scrypt"
  | "section"
  | "step"
  | "types" -> (identifier_prefix^name)
  | _ -> name
;;

(*
  We have used a hashtable to store the reserved words 
  (see~\ref{lexer:keyword:table}). First, we look in this
  table whether the parsed word is a reserved keyword. Then, 
  if this not the case, we associate the word to either a 
  variable or a constant following the prolog-convention of
  an uppercase first letter for variables.
*)
let handle_string identifier = 
  if is_keyword identifier then
    get_keyword identifier
  else
    CONST_IDENT(prefix_name identifier);;
}

(*
  We now parse the file. These rules define what tokens will be 
  recognized for the grammar.
*)

rule token = parse

(*
  First, we skip white spaces. For future, in the case of a newline,
  we keep track of a line number.
*)
    [' ' '\t' '\r']    {token lexbuf}
  | '\n'          {incr Error_handler.lineCounter;
		   Error_handler.offsetCounter := Lexing.lexeme_end lexbuf;
		   token lexbuf}

(*
  Then, we skip comments. Comments extend from a '\%' till the end of
  the line.
*)
  | '%' [^ '\n']* {token lexbuf}

(*
  Finally, for each other lexeme, we associate a token.
*) 
  | ['0'-'9']+    {NAT_IDENT(Lexing.lexeme lexbuf)}
  | "--|>"        {ACTION_ARROW}
  | "=|>"         {REACTION_ARROW}
  | "->"          {FUNCTION_ARROW}
  | "=>"          {IMPLIES}
  | ":="          {ASSIGN}
  | "def="        {DEF_EQUAL}
  | "<="          {LEQ}
  | "/="          {NEQ}
  | "/\\"         {AND}
  | "&"           {AND}
  | "\\/"         {OR}
  | "[]"          {LTL_op1}
  | "<->"         {LTL_op2}
  | "(-)"         {LTL_op3}
  | "[-]"         {LTL_op4}
  | "~"           {LTL_op5}
  | "("           {L_PAR}
  | ")"           {R_PAR}
(*
  | "["           {L_HOOK}
  | "]"           {R_HOOK}
*)
  | "{"           {L_BRACE}
  | "}"           {R_BRACE}
  | ","           {COMMA}
  | ":"           {COLON}
  | ";"           {SEMI_COLON}
  | "."           {DOT}
  | "+"           {PLUS}
  | "="           {EQUAL}
  | "'"           {PRIME}
  | "_"           {UNDERSCORE}
  | ['A'-'Z'] ['A'-'Z' 'a'-'z' '0'-'9' '_']* 
                  {VAR_IDENT (Lexing.lexeme lexbuf)}
  | ['a'-'z'] ['A'-'Z' 'a'-'z' '0'-'9' '_']*
                  {handle_string (Lexing.lexeme lexbuf)}
(*
  | "\"" [^ '\n' '\r' '\"' ]* "\""
                  {STRING(Lexing.lexeme lexbuf)}
*)
  | eof {raise Eof}
(* 
 | _             {Error_handler.handleLexerError 
		     (Lexing.lexeme lexbuf) 
		     " unknown symbol"; 
		   exit (-1)} 
*)
