
(* SPAN - A Security Protocol Animator
   Copyright (C) 2006-2007 Yann Glouche and Thomas Genet

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU Library General Public License version 2, as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU Library General Public License version 2 for more details
   (enclosed in the file LGPL).


   Version 1.0. Last modification 1/03/07
*)
(* ..................................................................
    Module ......: Unifyxor
    Initial Date : 03 January 2005
    Last Update .: 06 March 2006
    Developer ...: Max Tuengerthal
    Compiler ....: OCAML 3.07
    Comments ....: Unification algorithm for Xor/ACUN theory
    Project .....: The CL-Atse Protocol Analyser, AVISPA.
 .................................................................... *)

open Types_unify
open Utils_unify


(* generates and returns a default row ordering of length n,
 * i.e. this array: [|0; 1; 2; ...; n-2; n-1|]
 *)
let generate_row_ordering (n : int) : int array  =
  Array.init n (fun x -> x)
;;


(* returns the value of the entry at position (row,col) in the
 * matrix "matrix" with regard to the row ordering "ro"
 *)
let lookup ro (matrix : 'a array array) row col =
  matrix.(ro.(row)).(col)
;;


(* returns the row number "row" (with regard to the row ordering) of matrix
 *)
let get_row ro (matrix : 'a array array) row =
  matrix.(ro.(row))
;;


(* swaps line m and n, not physically, but by swapping the entries in the
 * row ordering "ro"
 *
 * Side effects: manipulates "ro"
 *)
let swap_lines ro m n =
  let buf = ro.(m) in
  ro.(m) <- ro.(n); ro.(n) <- buf
;;


(* adds (mod 2) row "m" and "n" of the matrices "a" and "b" and stores
 * the result in row "m"
 *)
let add_lines ro a b m n =
  let a_line_m = a.(ro.(m)) in
  let a_line_n = a.(ro.(n)) in
  let b_line_m = b.(ro.(m)) in
  let b_line_n = b.(ro.(n)) in
  for i = 0 to (Array.length a_line_m)-1 do
    a_line_m.(i) <- (a_line_m.(i) + a_line_n.(i)) mod 2
  done;
  for i = 0 to (Array.length b_line_m)-1 do
    b_line_m.(i) <- (b_line_m.(i) + b_line_n.(i)) mod 2
  done;;


(* looks for a row below row "start_row" with a "1" in column "col"
 * and swaps row "start_row" with the one found and returns "true"
 * in that case. In the other case it returns "false"
 *
 * Side effects: manipulates "ro"
 *)
let rec pivot ro matrix start_row row col =
  if lookup ro matrix row col = 1 then
    begin
      if start_row <> row then
        swap_lines ro start_row row;
      true
    end
  else
    begin
      if row < (Array.length matrix)-1 then
        pivot ro matrix start_row (row+1) col
      else
        false
    end;;


(* the system
 *
 * 1. a11*x + a12*y + a13*z = b11*e + b12*f + b13*g
 * 2. a21*x + a22*y + a23*z = b21*e + b22*f + b23*g
 * 3. a31*x + a32*y + a33*z = b31*e + b32*f + b33*g
 *
 * in which
 * - aij and bij are 0 or 1,
 * - x,y,z are variables and
 * - e,f,g are constants
 *
 * is represented by the two matrices a and b:
 *
 * A = [[a11 a12 a13]     B = [[b11 b12 b13]
 *      [a21 a22 a23]          [b21 b22 b23]
 *      [a31 a32 a33]]         [b31 b32 b33]]
 *
 * This algorithm turns this system (A,B) into an equivalent
 * system (A',B') with a triangular matrix A'
 *
 * Hypothesis:
 * - all entries in "a" and "b" are either 0 or 1
 * - both matrices are rectangular
 * - the height of both matrices is the same
 *
 * Output:
 * - an array "ml" that gives the ordering of the rows in the output matrices
 *
 * Side effects: modifies "a" and "b"
 *)
let gauss_elim (a : int array array) (b : int array array) =
  let height  = Array.length(a)     in
  let ro      = generate_row_ordering height  in
  if height=0 then
    ro
  else
    let row = ref 0
    in
    for col = 0 to (min (Array.length(a.(0))) height)-1 do
      (* do pivot search and test if it was successful *)
      if pivot ro a !row !row col then
        begin
          (* pivot search was succesfull, otherwise all elements in this column
           * are 0 and nothing has to be done *)
          for row2 = !row+1 to height-1 do
            (* eliminate 1's in column "col", i.e. add line !row to every line
             * with a 1 in col "col", except to itself *)
            if lookup ro a row2 col = 1 then
              add_lines ro a b row2 !row
          done;
          (* proceed to next row *)
          row := !row + 1
        end
    done;
    (* return the row ordering "ro" *)
    ro
;;


(* adds the list of variables in eqn::eqns to vars, but not the ones in atms and
 * no duplicates,
 * i.e. if vars_eqns is the set of variables in eqns then the output of
 * "get_vars atms vars eqns" is "(vars_eqns \ atms) union vars"
 *
 * Remarks: eqn::eqns must not contain any other term than Var(_)
 *)
let rec get_vars (atms : int list) (vars : int list)
  : term list list -> int list =

  let rec help vars = function
    | Var(n)::tail -> if List.mem n atms then
                        help vars tail
                      else
                        help (insert n vars) tail
    | []           -> vars
    | _            -> failwith "Unifyxor.get_vars: only Var(_) allowed here"
  in
  function
    | eqn::eqns -> get_vars atms (help vars eqn) eqns
    | []        -> vars
;;


(* Input:  eqnsys - a list of equations
 *         n      - the number of equations
 *                  (this has to hold: length(eqnsys) == n)
 *         vars   - an array of variables that are considered to be variables
 *         atms   - an array of variables that are considered to be atoms,
 *                  there must be no other variables in eqnsys than in vars and
 *                  atms,
 *                  vars and atms must be disjoint
 *
 * Output: (a,b)  - where a and b are matrices of height n,
 *                  a and b represent the given equation system eqnsys
 *
 * If eqnsys =
 *   [
 *     [x; y;          g];
 *     [x;    z; e;    g];
 *     [   y; z; e; f   ]
 *   ]
 * Then n has to be 3.
 * If vars = [|x; y; z|] and atms = [|e; f; g|] then this represents the
 * following system:
 *
 *   1. x + y     =?         g
 *   2. x +     z =? e +     g
 *   3.     y + z =? e + f
 *   (where + stands for xor)
 *
 * Then a and b will represent this system, too:
 *
 *       x y z            e f g
 * a = [[1 1 0]     b = [[0 0 1]
 *      [1 0 1]          [1 0 1]
 *      [0 1 1]]         [1 1 0]]
 *
 * So, "a * vars^t =? b * atms^t" ("vars^t" stands for the transposed "vars")
 * has the same meaning than the given system.
 *)
let build_matrices (eqnsys : term list list) n vars atms =
  let a = Array.make_matrix n (Array.length vars) 0 in
  let b = Array.make_matrix n (Array.length atms) 0 in
  let rec mem x = function
    | (Var y)::tail -> x=y || mem x tail
    | []            -> false
    | _             -> failwith ("Unifyxor.build_matrices: "^
                                 "only Var(_) allowed here")
  in
  let rec f matrix row ns eqnsys =
    match eqnsys with
    | eqn::tail ->
        for col = 0 to (Array.length ns)-1 do
          if mem ns.(col) eqn then
            (matrix.(row)).(col) <- 1
        done;
        f matrix (row+1) ns tail
    | [] -> ()
  in
  f a 0 vars eqnsys;
  f b 0 atms eqnsys;
  (a, b)
;;


(* Input: array  - an array with entries 0 or 1
 *        except - an index where array[except] is treated to be 0,
 *                 although it is 1
 *        func   - a function
 *        tas    - the output will be added to tas
 *
 * Output: a list that contains all elements of tas and
 *         for all i: func(i), if array[i]==1 and i<>expect
 *
 * Example: if array = [|0; 1; 0; 1; 1|], tas = [x; y] then
 *
 *          array_to_list func array tas 1 = [func(4); func(3); x; y]
 *
 *          (func(1) is omitted, because of except=1)
 *)
let array_to_list (func : int -> 'a) array (tas : 'a list) (except : int) =
  let result = ref tas in
  for i = 0 to (Array.length array)-1 do
    if i <> except && array.(i) = 1 then
      result := (func i)::!result
  done;
  !result
;;


(* Input: po    - a partial order, with the meaning:
 *                if x < y then y must not be a subterm of x
 *        b_row - an array with 0's or 1's
 *        atms  - an array of variable indices with the same length than b_row
 *        x     - a variable index
 *
 * Every 1 in b_row stands for a variable in atms. If b_row[i]=1 then b_row
 * contains, so to say, atms[i].
 *
 * Output: true, iff every variable y in b_rows (to be more precise: for every y
 *         with atms[i]=y and b_row[i]=1) is allowed to be a subterm of x,
 *         i.e. not (x < y)
 *)
let is_possible (po : t_po) b_row atms x =
  let majors = get_majors po x in
  let i = ref 0 in
  let possible = ref true in
  while !i < (Array.length atms) && !possible do
    (* if this atom is used and the atom is bigger then x, i.e. the
     * atom is in the list of the majors of x, then return false *)
    if b_row.(!i) = 1 && List.mem atms.(!i) majors then
      possible := false
    else
      i := !i+1
  done;
  !possible
;;


(* an internal exception, used only in solve_triangular *)
exception Substitution_found of (int * term) list;;


(* Input: a    - a triangular matrix
 *        b    - a matrix of the same height as a
 *        ro   - a row ordering for a and b
 *        vars - an array of variables that can obtain a value
 *        atms - an array of variables that must not obtain a value, i.e. are
 *               treated as constants
 *        po   - a partial order, with the meaning
 *               if x < y then y must not be a subterm of x
 *
 * Descritpion:
 * - Solve the system given by a, b, vars and atms.
 *
 * - Try all possible solutions, until one is found, that is valid according to
 *   the partial order po.
 *
 * - If none is found raise the exception No_Solution.
 *
 * - The internal exception Substitution_found is used to jump out of the
 *   process that iterates all possible solutions.
 *)
let solve_triangular a b ro vars atms po =
  let rec help po subst row =
    let no_ones_found = ref true in
    for col = row to (Array.length vars)-1 do
      (* if we found a one we test if it
       * is possible to choose this variable according
       * to the partial order *)
      if lookup ro a row col = 1 then
        begin
          (if is_possible po (get_row ro b row) atms vars.(col) then
            begin
              (* extend substitution *)

              let ll = array_to_list (fun i -> Var vars.(i))
                  (get_row ro a row) [] col
              in
              (* add the variables in atms to ll *)
              let ll = array_to_list (fun i -> Var atms.(i))
                  (get_row ro b row) ll (-1)
              in
              (* do a small normalisation *)
              let term = match ll with
                         | []  -> Atm(0)
                         | [x] -> x
                         | _   -> Xor ll
              in
              let subst = (vars.(col), term)::subst
              in
              (* try to solve system with the next row *)
              if row = 0 then
                (* we found a solution and can stop iterating *)
                raise (Substitution_found subst)
              else
                (* try to solve system with the next row *)

                (* update the partial order *)
                let po = it_list
                  (fun po t ->
                    match t with
                    | Var(m) -> add_to_po po m vars.(col)
                    | _      -> failwith "solve_triangular: only Var(_) allowed"
                  ) po ll
                in
                (* apply substitution to the other equations *)
                for row2 = 0 to row-1 do
                  if lookup ro a row2 col = 1 then
                    add_lines ro a b row2 row
                done;
                help po subst (row-1)
            end
          else
            ()
          );
          no_ones_found := false
        end
      else
        ()
    done;
    (* if there are no variables in this row we can find a solution
     * iff there are also no constants in this row, i.e. if it is
     * a 0 = 0 equation *)
    if !no_ones_found then
      begin
        for i = 0 to (Array.length atms)-1 do
          if lookup ro b row i = 1 then
            raise No_Solution
        done;
        if row = 0 then
          raise (Substitution_found subst)
        else
          help po subst (row-1)
      end
  in
  try
    help po [] ((Array.length a)-1);
    raise No_Solution
  with
    Substitution_found subst -> subst
;;


(* Input: eqns - a given xor unification problem, the only terms allowed in eqns
 *               are Var(_)
 *        po   - a partial order
 *        atms - list of variables, that are treated as constants
 *
 * Output: A most general unifier for the xor unification problem eqns.
 *
 *         Or raises the exception No_Solution, if eqns is not unifiable.
 *
 * Remarks to the partial order:
 * - restrictions between the variables not in atms are ignored (because in the
 *   std. theory these variables are constants, so there are no restrictions
 *   between them as they always appear on the right hand side)
 *
 * - restrictions between variables in atms are irrelevant (because they will
 *   never appear on the left hand side)
 *
 * Remarks:
 * - the output is NOT normalized, i.e. there can be terms like
 *   Xor [Var 3; Var 2] (wrong order!) or
 *   (1,Xor [Var 2; ...]), (2,Xor [...]) (Variable 2 has to be replaced in
 *   (1,...)
 *
 * - The output works together with the partial order, i.e. if n<m in the
 *   partial order then there is no substitution (n,Xor [...m...]) BUT there can
 *   be substitutions like (n, Var m) or (m, Var n)
 *)
let unif_xor (eqns : term list list) (po : t_po) (atms : int list) : t_equal =
  let n = List.length eqns in
  if n=0 then
    []
  else
    let vars = get_vars atms [] eqns in
    let vars_array = Array.of_list vars in
    let atms_array = Array.of_list atms in
    let a, b = build_matrices eqns n vars_array atms_array in
    let ro = gauss_elim a b in
    (* solve_trianglar may raise the exception No_Solution *)
    solve_triangular a b ro vars_array atms_array po
;;
